@@ -2,7 +2,7 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2012  Tieto Poland
+ *  Copyright (C) 2012-2013  Tieto Poland
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -46,68 +46,103 @@
 #define NEARD_MANAGER_INTERFACE "org.neard.Manager"
 #define AGENT_INTERFACE "org.neard.HandoverAgent"
 #define AGENT_PATH "/org/bluez/neard_handover_agent"
+#define AGENT_CARRIER_TYPE "bluetooth"
 #define ERROR_INTERFACE "org.neard.HandoverAgent.Error"
 
 static guint watcher_id = 0;
-static gboolean agent_registered = FALSE;
-static gboolean agent_register_postpone = FALSE;
+static char *neard_service = NULL;
+static bool agent_register_postpone = false;
 
 /* For NFC mimetype limits max OOB EIR size */
 #define NFC_OOB_EIR_MAX UINT8_MAX
 
+enum cps {
+	CPS_ACTIVE,
+	CPS_INACTIVE,
+	CPS_ACTIVATING,
+	CPS_UNKNOWN,
+};
+
+struct oob_params {
+	bdaddr_t address;
+	uint32_t class;
+	char *name;
+	GSList *services;
+	uint8_t *hash;
+	uint8_t *randomizer;
+	uint8_t *pin;
+	int pin_len;
+	enum cps power_state;
+};
+
+static void free_oob_params(struct oob_params *params)
+{
+	g_slist_free_full(params->services, g_free);
+	g_free(params->name);
+	g_free(params->hash);
+	g_free(params->randomizer);
+	g_free(params->pin);
+}
+
 static DBusMessage *error_reply(DBusMessage *msg, int error)
 {
-	switch (error) {
-	case ENOTSUP:
-		return g_dbus_create_error(msg, ERROR_INTERFACE ".NotSupported",
-						"Operation is not supported");
+	const char *name;
 
-	case ENOENT:
-		return g_dbus_create_error(msg, ERROR_INTERFACE ".NoSuchDevice",
-							"No such device");
+	if (error == EINPROGRESS)
+		name = ERROR_INTERFACE ".InProgress";
+	else
+		name = ERROR_INTERFACE ".Failed";
 
-	case EINPROGRESS:
-		return g_dbus_create_error(msg, ERROR_INTERFACE ".InProgress",
-						"Operation already in progress");
-
-	case ENONET:
-		return g_dbus_create_error(msg, ERROR_INTERFACE ".Disabled",
-							"Device disabled");
-
-	default:
-		return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
-							"%s", strerror(error));
-	}
+	return g_dbus_create_error(msg, name , "%s", strerror(error));
 }
 
+static void register_agent(bool append_carrier);
+
 static void register_agent_cb(DBusPendingCall *call, void *user_data)
 {
 	DBusMessage *reply;
 	DBusError err;
+	static bool try_fallback = true;
 
 	reply = dbus_pending_call_steal_reply(call);
 
 	dbus_error_init(&err);
 	if (dbus_set_error_from_message(&err, reply)) {
-		error("neard manager replied with an error: %s, %s",
-						err.name, err.message);
+		if (g_str_equal(DBUS_ERROR_UNKNOWN_METHOD, err.name) &&
+				try_fallback) {
+			DBG("Register to neard failed, trying legacy way");
+
+			register_agent(false);
+			try_fallback = false;
+		} else {
+			error("neard manager replied with an error: %s, %s",
+							err.name, err.message);
+
+			g_dbus_unregister_interface(btd_get_dbus_connection(),
+						AGENT_PATH, AGENT_INTERFACE);
+			try_fallback = true;
+		}
+
 		dbus_error_free(&err);
 		dbus_message_unref(reply);
 
-		g_dbus_unregister_interface(btd_get_dbus_connection(),
-						AGENT_PATH, AGENT_INTERFACE);
 		return;
 	}
 
 	dbus_message_unref(reply);
-	agent_registered = TRUE;
+	neard_service = g_strdup(dbus_message_get_sender(reply));
+
+	try_fallback = true;
+
+	info("Registered as neard handover agent");
 }
 
-static void register_agent(void)
+static void register_agent(bool append_carrier)
 {
 	DBusMessage *message;
 	DBusPendingCall *call;
 	const char *path = AGENT_PATH;
+	const char *carrier = AGENT_CARRIER_TYPE;
 
 	message = dbus_message_new_method_call(NEARD_NAME, NEARD_PATH,
 			NEARD_MANAGER_INTERFACE, "RegisterHandoverAgent");
@@ -119,6 +154,10 @@ static void register_agent(void)
 	dbus_message_append_args(message, DBUS_TYPE_OBJECT_PATH, &path,
 							DBUS_TYPE_INVALID);
 
+	if (append_carrier)
+		dbus_message_append_args(message, DBUS_TYPE_STRING, &carrier,
+							DBUS_TYPE_INVALID);
+
 	if (!dbus_connection_send_with_reply(btd_get_dbus_connection(),
 							message, &call, -1)) {
 		dbus_message_unref(message);
@@ -136,8 +175,10 @@ static void unregister_agent(void)
 {
 	DBusMessage *message;
 	const char *path = AGENT_PATH;
+	const char *carrier = AGENT_CARRIER_TYPE;
 
-	agent_registered = FALSE;
+	g_free(neard_service);
+	neard_service = NULL;
 
 	message = dbus_message_new_method_call(NEARD_NAME, NEARD_PATH,
 			NEARD_MANAGER_INTERFACE, "UnregisterHandoverAgent");
@@ -150,6 +191,9 @@ static void unregister_agent(void)
 	dbus_message_append_args(message, DBUS_TYPE_OBJECT_PATH, &path,
 						DBUS_TYPE_INVALID);
 
+	dbus_message_append_args(message, DBUS_TYPE_STRING, &carrier,
+							DBUS_TYPE_INVALID);
+
 	if (!g_dbus_send_message(btd_get_dbus_connection(), message))
 		error("D-Bus send failed");
 
@@ -158,6 +202,19 @@ unregister:
 							AGENT_INTERFACE);
 }
 
+static void add_power_state(DBusMessageIter *dict, struct btd_adapter *adapter)
+{
+	const char *state;
+
+	if (btd_adapter_get_powered(adapter) &&
+					btd_adapter_get_connectable(adapter))
+		state = "active";
+	else
+		state = "inactive";
+
+	dict_append_entry(dict, "State", DBUS_TYPE_STRING, &state);
+}
+
 static DBusMessage *create_request_oob_reply(struct btd_adapter *adapter,
 						const uint8_t *hash,
 						const uint8_t *randomizer,
@@ -193,6 +250,8 @@ static DBusMessage *create_request_oob_reply(struct btd_adapter *adapter,
 
 	dict_append_array(&dict, "EIR", DBUS_TYPE_BYTE, &peir, len);
 
+	add_power_state(&dict, adapter);
+
 	dbus_message_iter_close_container(&iter, &dict);
 
 	return reply;
@@ -207,12 +266,12 @@ static void read_local_complete(struct btd_adapter *adapter,
 
 	DBG("");
 
-	if (!agent_registered) {
+	if (neard_service == NULL) {
 		dbus_message_unref(msg);
 
 		if (agent_register_postpone) {
-			agent_register_postpone = FALSE;
-			register_agent();
+			agent_register_postpone = false;
+			register_agent(true);
 		}
 
 		return;
@@ -239,12 +298,12 @@ static void bonding_complete(struct btd_adapter *adapter,
 
 	DBG("");
 
-	if (!agent_registered) {
+	if (neard_service == NULL) {
 		dbus_message_unref(msg);
 
 		if (agent_register_postpone) {
-			agent_register_postpone = FALSE;
-			register_agent();
+			agent_register_postpone = false;
+			register_agent(true);
 		}
 
 		return;
@@ -281,14 +340,9 @@ static int check_device(struct btd_device *device)
 	return 0;
 }
 
-/* returns 1 if action (pairing or reading local data) is not needed */
-static int process_eir(struct btd_adapter *adapter, uint8_t *eir, size_t size,
-							bdaddr_t *remote)
+static int process_eir(uint8_t *eir, size_t size, struct oob_params *remote)
 {
 	struct eir_data eir_data;
-	char remote_address[18];
-	struct btd_device *device;
-	int err;
 
 	DBG("size %zu", size);
 
@@ -297,48 +351,25 @@ static int process_eir(struct btd_adapter *adapter, uint8_t *eir, size_t size,
 	if (eir_parse_oob(&eir_data, eir, size) < 0)
 		return -EINVAL;
 
-	ba2str(&eir_data.addr, remote_address);
+	bacpy(&remote->address, &eir_data.addr);
 
-	DBG("hci%u remote:%s", btd_adapter_get_index(adapter), remote_address);
+	remote->class = eir_data.class;
 
-	device = adapter_get_device(adapter, &eir_data.addr, BDADDR_BREDR);
+	remote->name = eir_data.name;
+	eir_data.name = NULL;
 
-	err = check_device(device);
-	if (err < 0) {
-		eir_data_free(&eir_data);
-		return err;
-	}
+	remote->services = eir_data.services;
+	eir_data.services = NULL;
 
-	/* store OOB data */
-	if (eir_data.class != 0)
-		device_set_class(device, eir_data.class);
+	remote->hash = eir_data.hash;
+	eir_data.hash = NULL;
 
-	/* TODO handle incomplete name? */
-	if (eir_data.name) {
-		adapter_store_cached_name(adapter_get_address(adapter),
-					&eir_data.addr, eir_data.name);
-		device_set_name(device, eir_data.name);
-	}
-
-	if (eir_data.hash)
-		btd_adapter_add_remote_oob_data(adapter, &eir_data.addr,
-					eir_data.hash, eir_data.randomizer);
-
-	/* TODO handle UUIDs? */
-
-	if (remote)
-		bacpy(remote, &eir_data.addr);
-
-	/*
-	 * In RequestOOB reply append local hash and randomizer only if
-	 * received EIR also contained it.
-	 */
-	if (!remote && !eir_data.hash)
-		err = -EALREADY;
+	remote->randomizer = eir_data.randomizer;
+	eir_data.randomizer = NULL;
 
 	eir_data_free(&eir_data);
 
-	return err;
+	return 0;
 }
 
 /*
@@ -363,16 +394,8 @@ static int process_eir(struct btd_adapter *adapter, uint8_t *eir, size_t size,
  * N bytes - name
  */
 
-struct nokia_com_bt {
-	bdaddr_t address;
-	uint32_t cod;
-	uint8_t pin[16];
-	int pin_len;
-	char *name;
-};
-
 static int process_nokia_long (void *data, size_t size, uint8_t marker,
-						struct nokia_com_bt *nokia)
+						struct oob_params *remote)
 {
 	struct {
 		bdaddr_t address;
@@ -386,26 +409,26 @@ static int process_nokia_long (void *data, size_t size, uint8_t marker,
 		return -EINVAL;
 
 	/* address is not reverted */
-	baswap(&nokia->address, &n->address);
+	baswap(&remote->address, &n->address);
 
-	nokia->cod = n->class[0] | (n->class[1] << 8) | (n->class[2] << 16);
+	remote->class = n->class[0] | (n->class[1] << 8) | (n->class[2] << 16);
 
 	if (n->name_len > 0)
-		nokia->name = g_strndup((char *)n->name, n->name_len);
+		remote->name = g_strndup((char *)n->name, n->name_len);
 
 	if (marker == 0x01) {
-		memcpy(nokia->pin, n->authentication, 4);
-		nokia->pin_len = 4;
+		remote->pin = g_memdup(n->authentication, 4);
+		remote->pin_len = 4;
 	} else if (marker == 0x02) {
-		memcpy(nokia->pin, n->authentication, 16);
-		nokia->pin_len = 16;
+		remote->pin = g_memdup(n->authentication, 16);
+		remote->pin_len = 16;
 	}
 
 	return 0;
 }
 
 static int process_nokia_short (void *data, size_t size,
-						struct nokia_com_bt *nokia)
+						struct oob_params *remote)
 {
 	struct {
 		bdaddr_t address;
@@ -419,21 +442,21 @@ static int process_nokia_short (void *data, size_t size,
 		return -EINVAL;
 
 	/* address is not reverted */
-	baswap(&nokia->address, &n->address);
+	baswap(&remote->address, &n->address);
 
-	nokia->cod = n->class[0] | (n->class[1] << 8) | (n->class[2] << 16);
+	remote->class = n->class[0] | (n->class[1] << 8) | (n->class[2] << 16);
 
 	if (n->name_len > 0)
-		nokia->name = g_strndup((char *)n->name, n->name_len);
+		remote->name = g_strndup((char *)n->name, n->name_len);
 
-	memcpy(nokia->pin, n->authentication, 4);
-	nokia->pin_len = 4;
+	remote->pin = g_memdup(n->authentication, 4);
+	remote->pin_len = 4;
 
 	return 0;
 }
 
 static int process_nokia_extra_short (void *data, size_t size,
-						struct nokia_com_bt *nokia)
+						struct oob_params *remote)
 {
 	struct {
 		bdaddr_t address;
@@ -445,102 +468,59 @@ static int process_nokia_extra_short (void *data, size_t size,
 	if (size != sizeof(*n) + n->name_len)
 		return -EINVAL;
 
-	bacpy(&nokia->address, &n->address);
+	bacpy(&remote->address, &n->address);
 
-	nokia->cod = n->class[0] | (n->class[1] << 8) | (n->class[2] << 16);
+	remote->class = n->class[0] | (n->class[1] << 8) | (n->class[2] << 16);
 
 	if (n->name_len > 0)
-		nokia->name = g_strndup((char *)n->name, n->name_len);
+		remote->name = g_strndup((char *)n->name, n->name_len);
 
 	return 0;
 }
 
-/* returns 1 if pairing is not needed */
-static int process_nokia_com_bt(struct btd_adapter *adapter, void *data,
-						size_t size, bdaddr_t *remote)
+static int process_nokia_com_bt(uint8_t *data, size_t size,
+						struct oob_params *remote)
 {
-	uint8_t *marker;
-	struct nokia_com_bt nokia;
-	struct btd_device *device;
-	int ret;
-	char remote_address[18];
+	uint8_t marker;
 
-	/* Support this only for PushOOB */
-	if (!remote)
-		return -EOPNOTSUPP;
+	marker = *data++;
+	size--;
 
-	marker = data++;
-	size --;
+	DBG("marker: 0x%.2x  size: %zu", marker, size);
 
-	DBG("marker: 0x%.2x  size: %zu", *marker, size);
-
-	memset(&nokia, 0, sizeof(nokia));
-
-	switch (*marker) {
+	switch (marker) {
 	case 0x00:
 	case 0x01:
 	case 0x02:
-		ret = process_nokia_long(data, size, *marker, &nokia);
-		break;
+		return process_nokia_long(data, size, marker, remote);
 	case 0x10:
-		ret = process_nokia_extra_short(data, size, &nokia);
-		break;
+		return process_nokia_extra_short(data, size, remote);
 	case 0x24:
-		ret = process_nokia_short(data, size, &nokia);
-		break;
+		return process_nokia_short(data, size, remote);
 	default:
-		info("Not supported Nokia NFC extension (0x%.2x)", *marker);
-		ret = -EPROTONOSUPPORT;
-		break;
+		warn("Not supported Nokia NFC extension (0x%.2x)", marker);
+		return -EPROTONOSUPPORT;
 	}
-
-	if (ret < 0)
-		return ret;
-
-	ba2str(&nokia.address, remote_address);
-	DBG("hci%u remote:%s", btd_adapter_get_index(adapter), remote_address);
-
-	device = adapter_get_device(adapter, &nokia.address, BDADDR_BREDR);
-
-	ret = check_device(device);
-	if (ret != 0) {
-		g_free(nokia.name);
-		return ret;
-	}
-
-	DBG("hci%u remote:%s", btd_adapter_get_index(adapter), remote_address);
-
-	if (nokia.name) {
-		adapter_store_cached_name(adapter_get_address(adapter), remote,
-								nokia.name);
-		device_set_name(device, nokia.name);
-		g_free(nokia.name);
-	}
-
-	if (nokia.cod != 0)
-		device_set_class(device, nokia.cod);
-
-	if (nokia.pin_len > 0) {
-		/* TODO
-		 * Handle PIN, for now only discovery mode and 'common' PINs
-		 * that might be provided by agent will work correctly.
-		 */
-	}
-
-	bacpy(remote, &nokia.address);
-
-	return 0;
 }
 
-static int process_params(DBusMessage *msg, struct btd_adapter *adapter,
-							bdaddr_t *remote)
+static enum cps process_state(const char *state)
+{
+	if (strcasecmp(state, "active") == 0)
+		return CPS_ACTIVE;
+
+	if (strcasecmp(state, "activating") == 0)
+		return CPS_ACTIVATING;
+
+	if (strcasecmp(state, "inactive") == 0)
+		return CPS_INACTIVE;
+
+	return CPS_UNKNOWN;
+}
+
+static int process_message(DBusMessage *msg, struct oob_params *remote)
 {
 	DBusMessageIter iter;
 	DBusMessageIter dict;
-	DBusMessageIter value;
-	DBusMessageIter entry;
-	const char *key;
-	int type;
 
 	dbus_message_iter_init(msg, &iter);
 
@@ -549,46 +529,87 @@ static int process_params(DBusMessage *msg, struct btd_adapter *adapter,
 
 	dbus_message_iter_recurse(&iter, &dict);
 
-	type = dbus_message_iter_get_arg_type(&dict);
-	if (type != DBUS_TYPE_DICT_ENTRY) {
-		if (!remote && type == DBUS_TYPE_INVALID)
-			return -EALREADY;
+	while (dbus_message_iter_get_arg_type(&dict) == DBUS_TYPE_DICT_ENTRY) {
+		DBusMessageIter value;
+		DBusMessageIter entry;
+		const char *key;
 
-		return -EINVAL;
+		dbus_message_iter_recurse(&dict, &entry);
+
+		if (dbus_message_iter_get_arg_type(&entry) != DBUS_TYPE_STRING)
+			goto error;
+
+		dbus_message_iter_get_basic(&entry, &key);
+		dbus_message_iter_next(&entry);
+
+		dbus_message_iter_recurse(&entry, &value);
+
+		if (strcasecmp(key, "EIR") == 0) {
+			DBusMessageIter array;
+			uint8_t *eir;
+			int size;
+
+			/* nokia.com:bt and EIR should not be passed together */
+			if (bacmp(&remote->address, BDADDR_ANY) != 0)
+				goto error;
+
+			if (dbus_message_iter_get_arg_type(&value) !=
+					DBUS_TYPE_ARRAY)
+				goto error;
+
+			dbus_message_iter_recurse(&value, &array);
+			dbus_message_iter_get_fixed_array(&array, &eir, &size);
+
+			if (process_eir(eir, size, remote) < 0)
+				goto error;
+		} else if (strcasecmp(key, "nokia.com:bt") == 0) {
+			DBusMessageIter array;
+			uint8_t *data;
+			int size;
+
+			/* nokia.com:bt and EIR should not be passed together */
+			if (bacmp(&remote->address, BDADDR_ANY) != 0)
+				goto error;
+
+			if (dbus_message_iter_get_arg_type(&value) !=
+					DBUS_TYPE_ARRAY)
+				goto error;
+
+			dbus_message_iter_recurse(&value, &array);
+			dbus_message_iter_get_fixed_array(&array, &data, &size);
+
+			if (process_nokia_com_bt(data, size, remote))
+				goto error;
+		} else if (strcasecmp(key, "State") == 0) {
+			DBusMessageIter array;
+			const char *state;
+
+			if (dbus_message_iter_get_arg_type(&value) !=
+					DBUS_TYPE_STRING)
+				goto error;
+
+			dbus_message_iter_recurse(&value, &array);
+			dbus_message_iter_get_basic(&value, &state);
+
+			remote->power_state = process_state(state);
+			if (remote->power_state == CPS_UNKNOWN)
+				goto error;
+		}
+
+		dbus_message_iter_next(&dict);
 	}
 
-	dbus_message_iter_recurse(&dict, &entry);
-
-	if (dbus_message_iter_get_arg_type(&entry) != DBUS_TYPE_STRING)
+	/* Check if 'State' was passed along with one of other fields */
+	if (remote->power_state != CPS_UNKNOWN
+			&& bacmp(&remote->address, BDADDR_ANY) == 0)
 		return -EINVAL;
 
-	dbus_message_iter_get_basic(&entry, &key);
-	dbus_message_iter_next(&entry);
+	return 0;
 
-	dbus_message_iter_recurse(&entry, &value);
-
-	/* All keys have byte array type values */
-	if (dbus_message_iter_get_arg_type(&value) != DBUS_TYPE_ARRAY)
-		return -EINVAL;
-
-	if (strcasecmp(key, "EIR") == 0) {
-		DBusMessageIter array;
-		uint8_t *eir;
-		int size;
-
-		dbus_message_iter_recurse(&value, &array);
-		dbus_message_iter_get_fixed_array(&array, &eir, &size);
-
-		return process_eir(adapter, eir, size, remote);
-	} else if (strcasecmp(key, "nokia.com:bt") == 0) {
-		DBusMessageIter array;
-		uint8_t *data;
-		int size;
-
-		dbus_message_iter_recurse(&value, &array);
-		dbus_message_iter_get_fixed_array(&array, &data, &size);
-
-		return process_nokia_com_bt(adapter, data, size, remote);
+error:
+	if (bacmp(&remote->address, BDADDR_ANY) != 0) {
+		free_oob_params(remote);
+		memset(remote, 0, sizeof(*remote));
 	}
 
 	return -EINVAL;
@@ -602,24 +623,52 @@ static int check_adapter(struct btd_adapter *adapter)
 	if (btd_adapter_check_oob_handler(adapter))
 		return -EINPROGRESS;
 
-	if (!btd_adapter_get_pairable(adapter))
-		return -ENONET;
-
 	if (!btd_adapter_ssp_enabled(adapter))
 		return -ENOTSUP;
 
 	return 0;
 }
 
+static void store_params(struct btd_adapter *adapter, struct btd_device *device,
+						struct oob_params *params)
+{
+	if (params->class != 0)
+		device_set_class(device, params->class);
+
+	if (params->name) {
+		adapter_store_cached_name(adapter_get_address(adapter),
+						&params->address, params->name);
+		device_set_name(device, params->name);
+	}
+
+	/* TODO handle UUIDs? */
+
+	if (params->hash) {
+		btd_adapter_add_remote_oob_data(adapter, &params->address,
+							params->hash,
+							params->randomizer);
+	} else if (params->pin_len) {
+		/* TODO
+		 * Handle PIN, for now only discovery mode and 'common' PINs
+		 * that might be provided by agent will work correctly.
+		 */
+	}
+}
+
 static DBusMessage *push_oob(DBusConnection *conn, DBusMessage *msg, void *data)
 {
 	struct btd_adapter *adapter;
 	struct agent *agent;
 	struct oob_handler *handler;
-	bdaddr_t remote;
+	struct oob_params remote;
+	struct btd_device *device;
 	uint8_t io_cap;
 	int err;
 
+	if (neard_service == NULL ||
+			!g_str_equal(neard_service, dbus_message_get_sender(msg)))
+		return error_reply(msg, EPERM);
+
 	DBG("");
 
 	adapter = btd_adapter_get_default();
@@ -628,14 +677,8 @@ static DBusMessage *push_oob(DBusConnection *conn, DBusMessage *msg, void *data)
 	if (err < 0)
 		return error_reply(msg, -err);
 
-	err = process_params(msg, adapter, &remote);
-
-	/* already paired, reply immediately */
-	if (err == -EALREADY)
-		return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
-
-	if (err < 0)
-		return error_reply(msg, -err);
+	if (!btd_adapter_get_powered(adapter))
+		return error_reply(msg, ENONET);
 
 	agent = adapter_get_agent(adapter);
 	if (!agent)
@@ -644,13 +687,49 @@ static DBusMessage *push_oob(DBusConnection *conn, DBusMessage *msg, void *data)
 	io_cap = agent_get_io_capability(agent);
 	agent_unref(agent);
 
-	err = adapter_create_bonding(adapter, &remote, BDADDR_BREDR, io_cap);
+	memset(&remote, 0, sizeof(remote));
+
+	err = process_message(msg, &remote);
+	if (err < 0)
+		return error_reply(msg, -err);
+
+	if (bacmp(&remote.address, BDADDR_ANY) == 0) {
+		free_oob_params(&remote);
+
+		return error_reply(msg, EINVAL);
+	}
+
+	device = adapter_get_device(adapter, &remote.address, BDADDR_BREDR);
+
+	err = check_device(device);
+	if (err < 0) {
+		free_oob_params(&remote);
+
+		/* already paired, reply immediately */
+		if (err == -EALREADY)
+			return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+
+		return error_reply(msg, -err);
+	}
+
+	if (!btd_adapter_get_pairable(adapter)) {
+		free_oob_params(&remote);
+
+		return error_reply(msg, ENONET);
+	}
+
+	store_params(adapter, device, &remote);
+
+	free_oob_params(&remote);
+
+	err = adapter_create_bonding(adapter, device_get_address(device),
+							BDADDR_BREDR, io_cap);
 	if (err < 0)
 		return error_reply(msg, -err);
 
 	handler = g_new0(struct oob_handler, 1);
 	handler->bonding_cb = bonding_complete;
-	bacpy(&handler->remote_addr, &remote);
+	bacpy(&handler->remote_addr, device_get_address(device));
 	handler->user_data = dbus_message_ref(msg);
 
 	btd_adapter_set_oob_handler(adapter, handler);
@@ -663,8 +742,14 @@ static DBusMessage *request_oob(DBusConnection *conn, DBusMessage *msg,
 {
 	struct btd_adapter *adapter;
 	struct oob_handler *handler;
+	struct oob_params remote;
+	struct btd_device *device;
 	int err;
 
+	if (neard_service == NULL ||
+			!g_str_equal(neard_service, dbus_message_get_sender(msg)))
+		return error_reply(msg, EPERM);
+
 	DBG("");
 
 	adapter = btd_adapter_get_default();
@@ -673,13 +758,44 @@ static DBusMessage *request_oob(DBusConnection *conn, DBusMessage *msg,
 	if (err < 0)
 		return error_reply(msg, -err);
 
-	err = process_params(msg, adapter, NULL);
-	if (err == -EALREADY)
-		return create_request_oob_reply(adapter, NULL, NULL, msg);
+	memset(&remote, 0, sizeof(remote));
 
+	err = process_message(msg, &remote);
 	if (err < 0)
 		return error_reply(msg, -err);
 
+	if (bacmp(&remote.address, BDADDR_ANY) == 0)
+		goto read_local;
+
+	device = adapter_get_device(adapter, &remote.address, BDADDR_BREDR);
+
+	err = check_device(device);
+	if (err < 0) {
+		free_oob_params(&remote);
+
+		if (err == -EALREADY)
+			return create_request_oob_reply(adapter, NULL, NULL,
+									msg);
+
+		return error_reply(msg, -err);
+	}
+
+	if (!btd_adapter_get_pairable(adapter)) {
+		free_oob_params(&remote);
+
+		return error_reply(msg, ENONET);
+	}
+
+	store_params(adapter, device, &remote);
+
+	if (!remote.hash) {
+		free_oob_params(&remote);
+		return create_request_oob_reply(adapter, NULL, NULL, msg);
+	}
+
+read_local:
+	free_oob_params(&remote);
+
 	err = btd_adapter_read_local_oob_data(adapter);
 	if (err < 0)
 		return error_reply(msg, -err);
@@ -696,9 +812,15 @@ static DBusMessage *request_oob(DBusConnection *conn, DBusMessage *msg,
 static DBusMessage *release(DBusConnection *conn, DBusMessage *msg,
 							void *user_data)
 {
+	if (neard_service == NULL ||
+			!g_str_equal(neard_service, dbus_message_get_sender(msg)))
+		return error_reply(msg, EPERM);
+
 	DBG("");
 
-	agent_registered = FALSE;
+	g_free(neard_service);
+	neard_service = NULL;
+
 	g_dbus_unregister_interface(conn, AGENT_PATH, AGENT_INTERFACE);
 
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
@@ -735,9 +857,9 @@ static void neard_appeared(DBusConnection *conn, void *user_data)
 	adapter = btd_adapter_get_default();
 
 	if (adapter && btd_adapter_check_oob_handler(adapter))
-		agent_register_postpone = TRUE;
+		agent_register_postpone = true;
 	else
-		register_agent();
+		register_agent(true);
 }
 
 static void neard_vanished(DBusConnection *conn, void *user_data)
@@ -745,8 +867,10 @@ static void neard_vanished(DBusConnection *conn, void *user_data)
 	DBG("");
 
 	/* neard existed without unregistering agent */
-	if (agent_registered) {
-		agent_registered = FALSE;
+	if (neard_service != NULL) {
+		g_free(neard_service);
+		neard_service = NULL;
+
 		g_dbus_unregister_interface(conn, AGENT_PATH, AGENT_INTERFACE);
 	}
 }
@@ -771,7 +895,7 @@ static void neard_exit(void)
 	g_dbus_remove_watch(btd_get_dbus_connection(), watcher_id);
 	watcher_id = 0;
 
-	if (agent_registered)
+	if (neard_service != NULL)
 		unregister_agent();
 }
 
